#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "global.h"
#include "rsaref.h"
#include "ripem.h"
#include "certder.h"
#include "keyderpr.h"
#include "keyfield.h"
#include "pubinfop.h"
#include "certutil.h"
#include "keymanpr.h"
#include "p.h"

/* For a Macintosh, where time is returned in seconds since
     1/1/1904 12:00:00 AM, YEAR_BASE should be defined as 1904.
   For Microsoft C 7.0, YEAR_BASE should be defined as 1900.
   The following defines YEAR_BASE as 1970 if it has not already been
     defined as something else with C compiler flags.
 */
#ifndef YEAR_BASE
#define YEAR_BASE 1970
#endif

/* Calculate the time adjustment the number of seconds between the year
     base and 1970. */

#ifdef MACTC
extern long GMTTimeshift(void);
#else
#define TIME_ADJUSTMENT \
  ((UINT4)(365 * (1970 - YEAR_BASE) + ((1970 - YEAR_BASE) + 2) / 4) * \
   (UINT4)24 * (UINT4)3600)

/* If the time() function returns local time, define GMT_OFFSET to be
     the number of hours that local time is EARLIER than GMT.
*/
#ifndef GMT_OFFSET
#define GMT_OFFSET 0
#endif
#endif  /* MACTC */

#define SECONDS_IN_MONTH \
  ((UINT4)((UINT4)365 * (UINT4)24 * (UINT4)3600) / (UINT4)12)

static void SignCert
  P((unsigned char *, unsigned int *, struct CertificateStruct *,
     R_RSA_PRIVATE_KEY *));

void R_time (theTime)
UINT4 *theTime;
{
  time ((time_t *)theTime);
#ifdef MACTC
  (*theTime) += GMTTimeshift();
#else
  /* Correct for a year base different than 1970 */
  (*theTime) -= TIME_ADJUSTMENT;

  /* Correct for local time to GMT */
  (*theTime) += (UINT4)3600 * GMT_OFFSET;
#endif
}

/* Create a self-signed cert with the userDN and publicKey which are
     in ripemInfo and sign it with ripemInfo's private key.  Write it
     to pubKeyOutFilename.
   Set the userCertDER in ripemInfo to the new self-signed cert.
 */
char *WriteSelfSignedCert (ripemInfo, validityMonths, pubKeyOutFilename)
RIPEMInfo *ripemInfo;
unsigned int validityMonths;
char *pubKeyOutFilename;
{
  CertificateStruct *certStruct = (CertificateStruct *)NULL;
  CertFieldPointers fieldPointers;
  unsigned int digestLen, maxCertDERLen;
  char *errorMessage = (char *)NULL;
  FILE *pubOutStream = (FILE *)NULL;

  /* For error, break to end of do while (0) block. */
  do {
    /* Open the output file.
     */
    pubOutStream = fopen(pubKeyOutFilename,"a");
    if(!pubOutStream) {
      sprintf(ripemInfo->errMsgTxt,"Can't open public key output file %s.",
              pubKeyOutFilename);
      errorMessage = ripemInfo->errMsgTxt;
      break;
    }
  
    /* Allocate the certStruct on the heap since it is so big. */
    if ((certStruct = (CertificateStruct *)malloc (sizeof (*certStruct)))
        == (CertificateStruct *)NULL) {
      errorMessage = ERR_MALLOC;
      break;
    }

    /* We are going to digest the struct to get a uniqe serial number, so
         pre-zerioze.
       This also pre-sets AVAIsT61 to 0. */
    R_memset ((POINTER)certStruct, 0, sizeof (*certStruct));

    certStruct->digestAlgorithm = DA_MD2;

    /* Set validity to now to now plus validity months. */
    R_time (&certStruct->notBefore);
    certStruct->notAfter = certStruct->notBefore +
      ((UINT4)((UINT4)validityMonths * SECONDS_IN_MONTH));

    /* Set subject name and issuer name to the user's DN. */
    certStruct->issuer = ripemInfo->userDN;
    certStruct->subject = certStruct->issuer;

    certStruct->publicKey = ripemInfo->publicKey;

    /* Now set the serial number to the digest of the certStruct-> */
    R_DigestBlock
      (certStruct->serialNumber, &digestLen, (unsigned char *)&certStruct,
       sizeof (certStruct), certStruct->digestAlgorithm);

    /* Allocate buffer for certificate DER and sign it.
       Alloca space for an extra byte as required by CodeAndWriteBytes.
     */
    maxCertDERLen =
      len_certificate (certStruct, PubKeyToDERLen (&certStruct->publicKey)) +
      4 + MAX_CERT_TO_SIGNED_DELTA + 1;

    if ((ripemInfo->z.userCertDER = (unsigned char *)malloc
         (maxCertDERLen)) == (unsigned char *)NULL)
      return (ERR_MALLOC);

    SignCert
      (ripemInfo->z.userCertDER, &ripemInfo->z.userCertDERLen, certStruct,
       &ripemInfo->privateKey);

    if ((errorMessage = WriteCert
         (ripemInfo->z.userCertDER, ripemInfo->z.userCertDERLen,
          pubOutStream)) != (char *)NULL)
      break;
  } while (0);

  if (pubOutStream != (FILE *)NULL)
    fclose (pubOutStream);
  free (certStruct);
  return (errorMessage);
}

/* Select the best certificate chain for the given distinguished name.
   If keyDigest is not NULL, digest the public key in the selected user
     certificate with digestAlgorithm and compare to keyDigest to make sure
     it is the right one.
   If keyDigest is NULL, ignore digestAlgorithm and allow any public key.
   The calling routine must initialize the certChain list.
   On return, certChain contains the chain where each entry is the
     certificate DER. The "bottom" of the chain comes first in the list.
     The caller can use DERToCertificate to decode the certs in the chain.
     Note that DERToCertificate will not return an error since it was already
       successfully decoded.
   This sets chainStatus to the values CERT_VALID, CERT_EXPIRED, etc.
     (chainStatus will not be set to CERT_UNVALIDATED.)
   Note: if no chain can be found, this returns a chainStatus of 0
     and the certChain is empty.
 */
char *SelectCertChain
  (ripemInfo, certChain, chainStatus, name, keyDigest, digestAlgorithm,
   ripemDatabase)
RIPEMInfo *ripemInfo;
TypList *certChain;
int *chainStatus;
DistinguishedNameStruct *name;
unsigned char *keyDigest;
int digestAlgorithm;
RIPEMDatabase *ripemDatabase;
{
  CertificateStruct *certStruct = (CertificateStruct *)NULL;
  CertFieldPointers fieldPointers;
  TypList certs;
  TypListEntry *entry;
  UINT4 now;
  char *errorMessage = (char *)NULL;
  int status, certLen;
  unsigned char localDigest[MAX_DIGEST_LEN];
  unsigned int digestLen, smartNameIndex;
  void *certCopy;

  InitList (&certs);

  /* For error, break to end of do while (0) block. */
  do {
    /* Get smart name of the sender to use in the search. */
    GetDNSmartNameIndex (&smartNameIndex, name);

    /* Select all certs for this username. */
    if ((errorMessage = GetCertsBySmartname
         (ripemDatabase, &certs, name->AVAValues[smartNameIndex], ripemInfo))
        != (char *)NULL)
      break;

    /* Allocate the certStruct on the heap since it is so big. */
    if ((certStruct = (CertificateStruct *)malloc
         (sizeof (*certStruct))) == (CertificateStruct *)NULL) {
      errorMessage = ERR_MALLOC;
      break;
    }
    
    /* Find the first cert with the ripemInfo's userDN as the issuer
         and matching key digest if given.
       Also, use the ripemInfo's public key to check the signature.
     */
    for (entry = certs.firstptr; entry; entry = entry->nextptr) { 
      if ((certLen = DERToCertificate
           ((unsigned char *)entry->dataptr, certStruct, &fieldPointers))
          < 0) {
        /* Error decoding.  Just issue a warning to debug stream and try
             the next cert. */
        if (ripemInfo->debug > 1)
          fprintf (ripemInfo->debugStream,
                   "Warning: Cannot decode certificate from database.\n");
        continue;
      }

      if (keyDigest != (unsigned char *)NULL) {
        R_DigestBlock
          (localDigest, &digestLen, fieldPointers.publicKeyDER,
           fieldPointers.publicKeyDERLen, digestAlgorithm);
        if (R_memcmp
            ((POINTER)keyDigest, (POINTER)localDigest, digestLen) != 0)
          /* not the correct public key */
          continue;
      }

      if (R_memcmp ((POINTER)&certStruct->issuer, (POINTER)&ripemInfo->userDN,
                    sizeof (certStruct->issuer)) != 0)
        /* The issuer is not the RIPEM user. */
        continue;

      if ((status = R_VerifyBlockSignature
           (fieldPointers.innerDER, fieldPointers.innerDERLen,
            certStruct->signature, certStruct->signatureLen,
            certStruct->digestAlgorithm, &ripemInfo->publicKey)) != 0) {
        /* Can't verify the signature.  Just print an error to the
             debug stream and try more certs. */
        if (ripemInfo->debug > 1)
          fprintf (ripemInfo->debugStream,
        "Warning: %s while checking signature on certificate from database.\n",
                   FormatRSAError (status));
        continue;
      }

      /* We have the correct certificate, so determine its status. */
      R_time (&now);
      if (certStruct->notBefore > now)
        *chainStatus = CERT_PENDING;
      else if (certStruct->notAfter < now)
        *chainStatus = CERT_EXPIRED;
      else
        *chainStatus = CERT_VALID;

      /* Make a copy of the certificate and prepend it to the certChain list.
         We prepend so that certificates at the "bottom" of the chain
           will be first.
       */
      if ((certCopy = malloc (certLen)) == NULL) {
        errorMessage = ERR_MALLOC;
        break;
      }
      R_memcpy ((POINTER)certCopy, (POINTER)entry->dataptr, certLen);
        
      if ((errorMessage = PrependToList (certCopy, certLen, certChain))
          != (char *)NULL) {
        /* Free the cert copy we just allocated */
        free (certCopy);
        break;
      }

      /* Finished */
      break;
    }
    if (errorMessage != (char *)NULL)
      break;

    if (!entry) {
      /* entry is NULL meaning that we looped through all the certs
          without finding one. */
      *chainStatus = 0;
      break;
    }
  } while (0);

  FreeList (&certs);
  free (certStruct);
  return (errorMessage);
}

/* On input, the certStruct has the name and public key to validate.
   This sets the issuerName to the user in ripemInfo, set the validity
     period to now plus validityMonths, and sets other fields to default
     values.  Then sign with ripemInfo's private key and insert in
     pubKeyOutStream.
   This is useful for validating the sender of a message after getting
     the self-signed certificate.
 */
char *ValidateAndWriteCert
  (ripemInfo, certStruct, validityMonths, pubKeyOutStream)
RIPEMInfo *ripemInfo;
CertificateStruct *certStruct;
unsigned int validityMonths;
FILE *pubKeyOutStream;
{
  unsigned char *certDER = (unsigned char *)NULL;
  unsigned int digestLen, maxCertDERLen, certDERLen;
  char *errorMessage = (char *)NULL;

  /* For error, break to end of do while (0) block. */
  do {
    certStruct->digestAlgorithm = DA_MD2;
    certStruct->issuer = ripemInfo->userDN;

    /* Set validity to now to now plus validity months. */
    R_time (&certStruct->notBefore);
    certStruct->notAfter = certStruct->notBefore +
      ((UINT4)((UINT4)validityMonths * SECONDS_IN_MONTH));

    /* Keep the same subject name and key. */

    /* Now set the serial number to the digest of the certStruct.
       It doesn't matter what this digest is, so long as it is unique. */
    R_DigestBlock
      (certStruct->serialNumber, &digestLen, (unsigned char *)&certStruct,
       sizeof (certStruct), certStruct->digestAlgorithm);

    /* Allocate buffer for certificate DER and sign it.
       Alloca space for an extra byte as required by CodeAndWriteBytes.
     */
    maxCertDERLen =
      len_certificate (certStruct, PubKeyToDERLen (&certStruct->publicKey)) +
      4 + MAX_CERT_TO_SIGNED_DELTA + 1;

    if ((certDER = (unsigned char *)malloc
         (maxCertDERLen)) == (unsigned char *)NULL) {
      errorMessage = ERR_MALLOC;
      break;
    }

    SignCert (certDER, &certDERLen, certStruct, &ripemInfo->privateKey);

    if ((errorMessage = WriteCert (certDER, certDERLen, pubKeyOutStream))
        != (char *)NULL)
      break;
  } while (0);

  free (certDER);
  return (errorMessage);
}

/* If the certificate issuer and subject are the same and the public
     key verifies the signature, set isSelfSigned to non-zero, otherwise
     set to zero.
 */
void CheckSelfSignedCert (isSelfSigned, certStruct, innerDER, innerDERLen)
int *isSelfSigned;
CertificateStruct *certStruct;
unsigned char *innerDER;
unsigned int innerDERLen;
{
  /* Default to not self-signed */
  *isSelfSigned = 0;

  if (R_memcmp ((POINTER)&certStruct->issuer, (POINTER)&certStruct->subject,
                sizeof (certStruct->issuer)) != 0)
    /* issuer != subject */
    return;

  if (R_VerifyBlockSignature
      (innerDER, innerDERLen, certStruct->signature,
       certStruct->signatureLen, certStruct->digestAlgorithm,
       &certStruct->publicKey) != 0)
    /* public key does not verify signature. */
    return;

  *isSelfSigned = 1;
}

/* Get the index in the AVAValues of the smart name.
 */
void GetDNSmartNameIndex (smartNameIndex, dn)
unsigned int *smartNameIndex;
DistinguishedNameStruct *dn;
{
  int typePriority = 0;
  unsigned int i;

  /* Go through the AVAs, setting nameIndex to the last common name, or
       to the last title if there are no common names.
   */
  for (i = 0; i < MAX_AVA; ++i) {
    if (dn->AVATypes[i] == -1)
      /* There are no more AVAs */
      break;

    if (dn->AVATypes[i] == ATTRTYPE_TITLE && typePriority <= 1) {
      *smartNameIndex = i;
      typePriority = 1;
    }
    if (dn->AVATypes[i] == ATTRTYPE_COMMONNAME && typePriority <= 2) {
      *smartNameIndex = i;
      typePriority = 2;
    }
  }

  if (typePriority == 0)
    /* There are no common names or titles, so use the least significant AVA */
    *smartNameIndex = i - 1;
}

/* certDER buffer is already allocated.  Encode and sign the info in certStruct
     with the privateKey and return the length in certDERLen.
   This leaves the signature in the certStruct buffer.
 */
static void SignCert (certDER, certDERLen, certStruct, privateKey)
unsigned char *certDER;
unsigned int *certDERLen;
CertificateStruct *certStruct;
R_RSA_PRIVATE_KEY *privateKey;
{
  CertificateToDer (certStruct, certDER, certDERLen);

  R_SignBlock
    (certStruct->signature, (unsigned int *)&certStruct->signatureLen, certDER,
     *certDERLen, DA_MD2, privateKey);

  DerCertToDerSigned
    (certDER, certDERLen, certStruct->signature, certStruct->signatureLen);
}

